/*
 * bmon - bandwidth monitor
 *
 * bmon is a bandwidth with 3 modes (overview, graphical, detailed). the
 * overview mode shows a list of all interfaces and the corresponding
 * byte/packet rates per second. The graphical mode displays the bandwidth
 * usage of the last 60 [seconds/minutes/hours] in a mrtg like format.
 * The detailed mode prints out values like errors, dropps, frames etc.
 *
 * Copyright (C) 2002 Thomas Graf <tgr@kaosu.ch>
 */


#include "bmon.h"
#include "intf.h"
#include "draw.h"
#include "conf.h"

static int         quit_done               = 0;
static int         brief_output_rem        = 1;
static int         signal_draw             = 0;
static int         initialized             = 0;
static int         do_html                 = 0;
static char        config_file[FILENAME_MAX];
char *             allowed_intf[255];
char *             denied_intf[255];


extern layout_t layout[];
extern int cfg_changed;
extern int cfg_use_colors;
extern int cfg_draw_interval;
extern char cfg_fill_char;
extern char cfg_bg_char;
extern int cfg_hide_dummy_interfaces;


#ifdef HAVE_CURSES
WINDOW *    win;
#endif

static char *usage_text =
"Usage: bmon [OPTION]...\n" \
"\n" \
"Options\n" \
"   -n              force using netstat even if /proc/net/dev is available\n" \
"   -a              show rates for interface aliases (netstat input method)\n" \
"   -p              plain output mode\n" \
"   -s              signal driven output mode (plain)\n" \
"   -H <path>       HTML output mode\n" \
"   -E              hide dummy interfaces sit0, irlan0, .. (linux)\n" \
"   -S[<pid>]       send USR1 signal to a running bmon\n" \
"   -d <name>       diagram type\n" \
"   -x <unit>       unit for x-axis (default is dynamic)\n" \
"   -y <unit>       unit for y-axis (default is dynamic)\n" \
"   -i <intf>       interface selection (default is all)\n" \
"   -f <char>       foreground char for graphical mode (default is '*')\n" \
"   -b <char>       background char for graphical mode (default is '.')\n" \
"   -D <seconds>    draw interval (default is eq. read interval)\n" \
"   -F <path>       alternative bmon configuration file path\n" \
"   -m              monochrome mode (no colors)\n" \
"   -h              show this help text\n" \
"   -V              show version\n" \
"\n" \
"Please see the bmon(1) man pages for full documentation.\n";


enum input_methods {
    proc_net_dev,
    netstat_i,
    netstat_i_a,
    kstat,
    bsd_sysctl
};

enum output_methods {
#ifdef HAVE_CURSES
    curses,
#endif
    brief,
    signal_driven
};

#if defined HAVE_KSTAT && defined SYS_SUNOS
static enum input_methods input_method = kstat;
#elif defined SYS_BSD
static enum input_methods input_method = bsd_sysctl;
#else
static enum input_methods input_method = netstat_i;
#endif

#ifdef HAVE_CURSES
static enum output_methods output_method = curses;
#else
static enum output_methods output_method = brief;
#endif

static char html_output_path[FILENAME_MAX];

extern enum bandwidth m_bandwidth;
extern enum graphic m_b_graphic;
extern enum unit m_unit;


RETSIGTYPE
finish(int unused)
{
    exit(unused);
}

RETSIGTYPE
sig_usr1(int unused)
{
    signal_draw = 1;
}


void
doexit(void)
{
    if (!quit_done && initialized)
    {
#ifdef HAVE_CURSES
        endwin();
#endif
        quit_done = 1;
    }
}

void
quit(const char *fmt, ...)
{
    va_list args;

    doexit();

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    exit(1);
}


void
p_parse_intf(const char *intf)
{
    int i, a=0, d=0, f=0;
    char *p, *s = strdup(intf);

    p = s;

    for (i=0; ; i++)
    {
        if (s[i] == ',' || s[i] == ' ' || s[i] == '\0')
        {
            if (s[i] == '\0')
                f = 1;

            s[i] = '\0';

            if (*p == '!')
            {
                if (d > 254)
                    break;

                denied_intf[d++] = strdup(++p);
            }
            else
            {
                if(a > 254)
                    break;

                allowed_intf[a++] = strdup(p);
            }

            if (f)
                break;
            else
                f = 0;

            p = &s[i+1];
        }
    }

    free(s);
}


void
print_version(void)
{

    printf("%s\n", PACKAGE_STRING);
    printf("Copyright (C) 2002 by Thomas Graf\n");
    printf("bmon comes with ABSOLUTELY NO WARRANTY. This is free software, " \
        "and you\nare welcome to redistribute it under certain conditions. " \
        "See the Artistic\nLicence for details.\n");
}


void
p_send_signal(const char *arg)
{
    if (arg)
    {
        long pid = strtol(arg, NULL, 10);
        if (pid != LONG_MIN)
        {
            if (kill((int) pid, SIGUSR1) < 0)
                quit("kill(%d, SIGUSR1) failed: %s\n",
                        (int) pid, strerror(errno));
        }
        else
            quit("Invalid pid.\n");
    }
    else
    {
        FILE *f;
        char buf[256], hdr[256], bmon_procs[10][256];
        int cnt=0, pid_to_kill=-1;
        int my_uid = getuid();
        char cmd[64];

        memset(cmd, 0, sizeof(cmd));
        snprintf(cmd, sizeof(cmd)-1, "ps -U %d -o pid,uid,comm,args",
                my_uid);

        if (!(f = popen(cmd, "r")))
        {
            fprintf(stderr, "popen(%s: %s\n", cmd, strerror(errno));
            quit("Your ps is probably not POSIX complaint, use kill\n");
        }

        fgets(hdr, sizeof(hdr), f);

        while(fgets(buf, sizeof(buf), f))
        {
            int pid, uid, n;
            char args[256], cmd[256];

            if (!strstr(buf, "bmon") ||
                !strstr(buf, "-s") ||
                 strstr(buf, "-S"))
            {
                continue;
            }

            n = sscanf(buf, "%d %d %s %[^\n\r]s",
                    &pid, &uid, cmd, args);

            if (n != 4)
                continue;

            if (uid == my_uid)
            {
                if (cnt <= 9)
                {
                    snprintf(bmon_procs[cnt], sizeof(bmon_procs[cnt])-1,
                            "%5d %5d %s %s", pid, uid, cmd, args);
                    pid_to_kill = pid;
                    cnt++;
                }
                else
                    break;
            }
        }

        pclose(f);

        if (cnt == 1)
        {
            if (pid_to_kill >= 0)
                if (kill((int) pid_to_kill, SIGUSR1) < 0)
                    quit("kill(%d, SIGUSR1) failed: %s\n",
                            (int) pid_to_kill, strerror(errno));
        }
        else
        {
            int n;

            fprintf(stderr, "Multiple instances of bmon are running, " \
                    "use bmon -S<pid>\n");

            printf("%s", hdr);
            for (n=0; n < cnt; n++)
                printf("%s\n", bmon_procs[n]);
        }
        
        finish(0);
    }
}



int
main(int argc, char *argv[])
{
    int c;

    atexit(&doexit);

    if (signal(SIGINT, finish) == SIG_ERR)
        quit("signal(SIGINT) failed: %s\n", strerror(errno));

    if (signal(SIGUSR1, sig_usr1) == SIG_ERR)
        quit("signal(SIGUSR1) failed: %s\n", strerror(errno));


    if (access("/proc/net/dev", R_OK) == 0)
        input_method = proc_net_dev;

    {
        char path[FILENAME_MAX];

        conf_read("/etc/bmon.conf");

        if (getenv("HOME"))
        {
            snprintf(path, sizeof(path), "%s/.bmonrc", getenv("HOME"));
            conf_read(path);
        }
    }

    memset(config_file, 0, sizeof(config_file));

    for (;;)
    {
        c = getopt(argc, argv, "anphVsd:x:y:i:f:b:S::H:D:EF:m");

        if (c == -1)
            break;
        
        switch (c)
        {
            case 'a':
                input_method = netstat_i_a;
                break;

            case 'n':
                input_method = netstat_i;
                break;

            case 'p':
                output_method = brief;
                break;

            case 's':
                output_method = signal_driven;
                break;

            case 'H':
                do_html = 1;
                memset(html_output_path, 0, sizeof(html_output_path));
                strncpy(html_output_path, optarg, sizeof(html_output_path)-1);
                break;

            case 'd':
                if (optarg[0] == 'o' || optarg[0] == 'O')
                    m_bandwidth = overview;
                else if (optarg[0] == 'd' || optarg[0] == 'D')
                    m_bandwidth = detailed;
                else if (optarg[0] == 'g' || optarg[0] == 'G')
                    m_bandwidth = graphic;
                else
                {
                    fprintf(stderr, "Unknown mode\n");
                    finish(1);
                }
                break;

            case 'x':
                if (optarg[0] == 's' || optarg[0] == 'S')
                    m_b_graphic = minute;
                else if (optarg[0] == 'm' || optarg[0] == 'M')
                    m_b_graphic = hour;
                else if (optarg[0] == 'h' || optarg[0] == 'H')
                    m_b_graphic = day;
                else
                {
                    fprintf(stderr, "Unknown X-axis unit\n");
                    finish(1);
                }
                break;

            case 'y':
                if (optarg[0] == 'b' || optarg[0] == 'B')
                    m_unit = byte;
                else if (optarg[0] == 'k' || optarg[0] == 'K')
                    m_unit = kilobyte;
                else if (optarg[0] == 'm' || optarg[0] == 'M')
                    m_unit = megabyte;
                else if (optarg[0] == 'g' || optarg[0] == 'G')
                    m_unit = gigabyte;
                else if (optarg[0] == 't' || optarg[0] == 'T')
                    m_unit = terabyte;
                else
                {
                    fprintf(stderr, "Unknonwn or not supported unit\n");
                    finish(1);
                }
                break;

            case 'i':
                p_parse_intf(optarg);
                break;

            case 'f':
                cfg_fill_char = optarg[0];
                cfg_changed |= CFG_FILL_CHAR;
                break;

            case 'b':
                cfg_bg_char = optarg[0];
                cfg_changed |= CFG_BG_CHAR;
                break;

            case 'h':
                print_version();
                printf("\n%s", usage_text);
                finish(0);
                break;

            case 'V':
                print_version();
                finish(0);
                break;

            case 'S':
                p_send_signal(optarg);
                break;

            case 'D':
                cfg_draw_interval = atoi(optarg);
                cfg_changed |= CFG_DRAW_INTERVAL;
                break;

            case 'E':
                cfg_hide_dummy_interfaces = 1;
                cfg_changed |= CFG_HIDE_DUMMY_INTERFACES;
                break;

            case 'F':
                strncpy(config_file, optarg, sizeof(config_file)-1);
                cfg_changed |= CFG_CONFIGFILE;
                break;

            case 'm':
                cfg_use_colors = 0;
                cfg_changed |= CFG_USE_COLORS;
                break;
        }
    }

    if (cfg_changed & CFG_CONFIGFILE)
        conf_read(config_file);

   
#ifdef HAVE_CURSES
    if (output_method == curses)
    {
        win = initscr();

        if (!win)
            exit(1);

        if (cfg_use_colors && has_colors())
        {
            int i;

            start_color();

#ifdef HAVE_USE_DEFAULT_COLORS
            use_default_colors();
#endif

            for (i = 1; i < NR_LAYOUTS+1; i++)
                init_pair(i, layout[i].fg, layout[i].bg);
        }
        else
            cfg_use_colors = 0;
        
        initialized = 1;

        keypad(stdscr, TRUE);
        nonl();
        cbreak();
        noecho();
        nodelay(win, TRUE);    /* getch etc. must be non-blocking */
        clear();
        curs_set(0);
    }
#endif

    
    /* main loop */
    {
        struct timeval    iread, idraw;
    
        gettimeofday(&iread, NULL);
        gettimeofday(&idraw, NULL);
        iread.tv_sec--;   /* hack to force reading the file for the first time */
        idraw.tv_sec -= cfg_draw_interval;


        for (;;)
        {
#ifdef HAVE_CURSES
            if (output_method == curses)
            {
                for (;;)
                {
                    int ch = getch();

                    if (ch == -1)
                        break;

                    if (ch == 'q')
                        finish(0);
                    else if (ch == 12 || ch == KEY_CLEAR)
                    {
                        clear();
                        intf_draw();
                    }
                    else if (intf_handle_input(ch))
                            intf_draw();
                }
            }
#endif


            if (p_diff_now(&iread) >= 1.0)
            {
                switch(input_method)
                {
                    case proc_net_dev:
                        intf_readfile("/proc/net/dev");
                        break;

                    case netstat_i:
                        intf_read_netstat("netstat -i -n");
                        break;

                    case netstat_i_a:
                        intf_read_netstat("netstat -i -a -n");
                        break;

                    case kstat:
#if defined HAVE_KSTAT && defined SYS_SUNOS
                        intf_read_kstat();
#endif
                        break;

                    case bsd_sysctl:
#if defined SYS_BSD
                        intf_read_sysctl();
#endif
                        break;
                }

                gettimeofday(&iread, NULL);
            }

            if (p_diff_now(&idraw) >= cfg_draw_interval)
            {
                switch (output_method)
                {
                    case brief:
                        if (brief_output_rem == 0)
                        {
                            intf_draw_ascii();
                            finish(0);
                        }
                        else
                            brief_output_rem--;
                        break;

#ifdef HAVE_CURSES
                    case curses:
                        intf_draw();
                        break;
#endif

                    case signal_driven:
                        if (signal_draw)
                        {
                            intf_draw_ascii();
                            signal_draw = 0;
                        }
                        break;

                    default:
                        finish(0);
                        break;
                }

                if (do_html)
                    intf_draw_html(html_output_path);

                gettimeofday(&idraw, NULL);
            }

            usleep(20000);
        }
    }

    return 0; /* will never be reached */
}
